/*
 * A quick-and-dirty data verification program
 * 
 * performs a write-and-verify pass over a file or block device.  Writes
 * 1MB at a time, and encodes the data offset in the block so that data
 * getting written to the wrong address can be detected.
 *
 * Defaults to verifying offsets every 1TB into the storage, but can
 * optionally test the entire contents.
 *
 * Copyright 2005 Red Hat, Inc.  All rights reserved.  This copyrighted
 * material is made available to anyone wishing to use, modify, copy, or
 * redistribute it subject to the terms and conditions of the GNU
 * General Public License v.2.
 *
 * Authors: Stephen C. Tweedie <sct@redhat.com>
 *
 * TODO:
 *  Add versioning/changelog
 *  Add immediate verify-after-write
 *  Handle raw IO devices on 2.4
 *  Handle filesystem EOF more gracefully
 *  Handle "start at this block number" option
 *  Add random seek mode
 *     Can we guarantee whole-device coverage for writes?
 *     Or should we restrict it to reads or immediate verify?
 */

/* We *really* want to be 64-bit throughout here */
#define _LARGEFILE_SOURCE
#define _LARGEFILE64_SOURCE

/* We want pread/pwrite and posix_memalign */
#define _XOPEN_SOURCE 600

#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <sys/ioctl.h>
#include <linux/fs.h>

/* Hack to be able to include both fcntl.h and linux/fcntl.h */
#define flock redefined_flock
#define flock64 redefined_flock64
/* For O_DIRECT: */
#include <linux/fcntl.h>
#undef flock

#define __stringify_1(x)	#x
#define __stringify(x)		__stringify_1(x)

static char *version = __stringify(VERSION);

struct buffer_pattern 
{
	loff_t	file_offset;
	loff_t	counter;
	loff_t	byte_offset;
	loff_t	magic;
};

#define PATTERN_MAGIC 0x01030507090B0D0FLL

char *progname;
char *filename;
loff_t default_step = (128 * 1024LL * 1024LL * 1024LL); /* 128GB */
int default_chunksize = (1024 * 1024);			/* 1MB */

loff_t file_limit = -1;

char *buffer;

int progress_pending = 0;
int progress_tty = 1;
int quiet = 0;

void progress(const char *format, ...)
{
	if (quiet)
		return;
	if (progress_tty) {
		if (progress_pending)
			putchar('\r');
		progress_pending = 1;
	}

	va_list ap;
	va_start(ap, format);
	vprintf(format, ap);
	va_end(ap);

	if (progress_tty) 
		fflush(stdout);
	else
		putchar('\n');
}

void flush_console(void)
{
	if (progress_pending)
		putchar('\n');
	fflush(stdout);
	progress_pending = 0;
}

void useage(int do_exit)
{
	fprintf(stderr, 
		"Useage: %s [-1ahQqrw] [-s iosize] filename [step] [cycles]\n",
		progname);
	if (do_exit)
		exit(1);
}

void show_help(void)
{
	useage(0);
	fprintf(stderr,
		"Performs a write/verify cycle over a given file or device.\n"
		"  -1:  perform one cycle only (skip to last cycle)\n"
		"  -a:  append to existing file (regular files only)\n"
		"  -h:  show this help\n"
		"  -V:  show program version\n"
		"  -Q:  query device size only, perform no IO\n"
		"  -q:  operate quietly\n"
		"  -r:  perform only read pass (verify existing data)\n"
		"  -w:  perform only write pass (write pattern for future verify)\n"
		"  -s:  set size of each IO (default 1MB)\n"
		"  filename:   specify the file or device to use\n"
		"  step:       specify distance between IOs submitted (default 128GB)\n"
		"  cycles:     specify how many complete cycles to repeat (default 1)\n"
		"  Suffixes [kKmMbBtT] may be used when specifying steps/size arguments\n"
		"  (all in multiples of 1024, no bogus decimal pseudo-metrics here.)\n"
		"  -r and -w are limited to a single cycle, but will use the given cycle\n"
		"  number in their data patterns.\n"
		);
}

void show_version(void)
{
	printf("%s version %s\n", progname, version);
}

loff_t memparse(const char *s)
{
	loff_t num;
	char *next;
	
	num = strtol(s, &next, 0);
	switch (*next) {
	case 't':
	case 'T':
		num *= 1024;
	case 'g':
	case 'G':
		num *= 1024;
	case 'm':
	case 'M':
		num *= 1024;
	case 'k':
	case 'K':
		num *= 1024;
		next++;
	}
	
	if (*next) {
		fprintf(stderr, "Unable to parse \"%s\" as a size\n", s);
		useage(1);
	}
	
	return num;
}

char obuf1[32], obuf2[32];
const char *describe_offset(loff_t offset, char *ubuf)
{
	static char buf[32];
	int i;
	loff_t unit = 1, offset_left;
	const char *units = "\0KMGTP";
	
	if (!ubuf)
		ubuf = buf;
	
	for (i = 0, unit = 1, offset_left = offset; 
	     (!i || units[i]) && offset_left >= 1024LL;
	     offset_left >>= 10, unit <<= 10, i++);
	
	snprintf(ubuf, 32, "%4.2f%c", 
		 ((double) offset) / ((double) unit), units[i]);
	return ubuf;
}

void die_perror(const char *format, ...) 
{
	flush_console();
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	fprintf(stderr, ": %s\n", strerror(errno));
	exit(1);
}

void die(const char *format, ...) 
{
	flush_console();
	va_list ap;
	va_start(ap, format);
	vfprintf(stderr, format, ap);
	va_end(ap);
	exit(1);
}

void query_state(int fd, loff_t *size, int verbose)
{
	struct stat64 statbuf;
	int rc, got_size = 0;
	const char *itype;
	unsigned long blocks;
	
	rc = fstat64(fd, &statbuf);
	if (rc < 0)
		die_perror("stat64");

	switch (statbuf.st_mode & S_IFMT) {
	case S_IFREG:
		itype = "regular file";
		*size = statbuf.st_size;
		break;
	case S_IFBLK:
		itype = "block device";
	treat_as_block_device:
#ifdef BLKGETSIZE64
		rc = ioctl(fd, BLKGETSIZE64, size);
		if (rc < 0)
			die_perror("ioctl(BLKGETSIZE64)");
		if (verbose)
			printf("BLKGETSIZE64 reports size is %jd bytes\n", 
			       *size);
		got_size++;
#endif

		if (verbose || !got_size) {
			rc = ioctl(fd, BLKGETSIZE, &blocks);
			/* We can't die on EFBIG here if BLKGETSIZE64
			 * already succeeded; BLKGETSIZE will fail for
			 * very large devices. */
			if (rc < 0) {
				if (errno == EFBIG && got_size)
					perror("ioctl(BLKGETSIZE)");
				else
					die_perror("ioctl(BLKGETSIZE)");
			} else {
				if (verbose)
					printf("BLKGETSIZE reports size is "
					       "%lu 512-byte blocks "
					       "(%jd bytes)\n",
					       blocks, 
					       ((loff_t) blocks) * 512);
				if (!got_size) {
					*size = (loff_t) blocks * 512;
					got_size++;
				}
			}
		}

		/* If we're in verbose mode, output a few more ioctls
		 * to show device state; we're not actually going to use 
		 * the information but it's useful for reporting. */
		if (verbose) {
			unsigned int sectsize;
			unsigned short sectlimit;
			long ra_pages;
			int is_rdonly;
			
			rc = ioctl(fd, BLKBSZGET, &sectsize);
			if (rc < 0)
				perror("ioctl(BLKBSZGET)");
			else
				printf("BLKBSZGET reports logical block size "
				       "%d bytes\n",
				       sectsize);
			
			rc = ioctl(fd, BLKSSZGET, &sectsize);
			if (rc < 0)
				perror("ioctl(BLKGETSIZE)");
			else
				printf("BLKSSZGET reports physical block size "
				       "%d bytes\n",
				       sectsize);

			rc = ioctl(fd, BLKRAGET, &ra_pages);
			if (rc < 0)
				perror("ioctl(BLKRAGET)");
			else
				printf("BLKRAGET reports %ld readahead pages\n",
				       ra_pages);

			rc = ioctl(fd, BLKROGET, &is_rdonly);
			if (rc < 0)
				perror("ioctl(BLKROGET)");
			else
				printf("BLKROGET reports device is %s\n",
				       is_rdonly ? "read-only" : "writable");

			rc = ioctl(fd, BLKSECTGET, &sectlimit);
			if (rc < 0)
				perror("ioctl(BLKSECTGET)");
			else
				printf("BLKSECTGET reports IO limit of "
				       "%d sectors\n",
				       (int) sectlimit);
		}
		
		break;

	case S_IFCHR:
		itype = "character device";
		/* It might be raw... let's hope */
		goto treat_as_block_device;
	default:
		fprintf(stderr, 
			"%s: File is not a regular file or block device\n",
			progname);
		exit(1);
	}
	printf("File is a %s: size %jd bytes (%s)\n", itype, 
	       *size, describe_offset(*size, 0));
}

void prepare_buffer(int chunksize, loff_t offset, int counter)
{
	struct buffer_pattern *p;
	int i, max;

	p = (struct buffer_pattern *) buffer;
	max = chunksize / sizeof(*p);
	
	for (i = 0; i < max; i++, p++) {
		p->file_offset = offset;
		p->counter = counter;
		p->byte_offset = i;
		p->magic = PATTERN_MAGIC;
	}
}

void complain(int *noerr, loff_t foffset, int boffset, const char *format, ...)
{
	flush_console();
	if (*noerr) {
		*noerr = 0;

		fprintf(stderr,
			"  Compare error at chunk %jd (%s), offset %d:\n",
			foffset, describe_offset(foffset, 0), boffset);
	}
	va_list ap;
	va_start(ap, format);
	fprintf(stderr, "    ");
	vfprintf(stderr, format, ap);
	va_end(ap);
}

int compare_buffer(int chunksize, loff_t offset, int counter)
{
	struct buffer_pattern *p;
	int i, max;
	int noerr = 1;
	
	p = (struct buffer_pattern *) buffer;
	max = chunksize / sizeof(*p);
	
	for (i = 0; i < max; i++, p++) {
		if (p->file_offset != offset)
			complain(&noerr, offset, i, 
				 "Expecting offset field %#zx, got %#zx\n",
				 offset, p->file_offset);
		if (p->counter != counter)
			complain(&noerr, offset, i, 
				 "Expecting cycle counter %d, got %d\n",
				 counter, p->counter);
		if (p->byte_offset != i)
			complain(&noerr, offset, i, 
				 "Expecting byte offset %#x, got %#x\n",
				 i, p->byte_offset);
		if (p->magic != PATTERN_MAGIC)
			complain(&noerr, offset, i, 
				 "Expecting magic %#llx, got %#llx\n",
				 PATTERN_MAGIC, p->magic);
		if (!noerr)
			return -1;
	}
	return 0;
}

/* Data read function:
 * return 0 on success,
 *       >0 on IO failure,
 *       <0 on (continuable) compare failure.
 */
int read_one(int fd, int chunksize, loff_t offset, int counter)
{
	int rc;
	rc = pread64(fd, buffer, chunksize, offset);
	if (rc < 0)
		return errno;
	if (rc != chunksize) 
		return EIO;
	return compare_buffer(chunksize, offset, counter);
}

/* Data write function:
 * return 0 on success,
 *       >0 on failure
 */
int write_one(int fd, int chunksize, loff_t offset, int counter)
{
	int rc;
	prepare_buffer(chunksize, offset, counter);
	rc = pwrite64(fd, buffer, chunksize, offset);
	if (rc < 0) {
		if (errno == ENOSPC || errno == EFBIG)
			if (file_limit < 0)
				file_limit = offset;
		return errno;
	}
	if (rc != chunksize) 
		return EIO;
	return 0;
}


int iterate_over_file(int fd, loff_t start, 
		      int chunksize, loff_t step, int counter,
		      const char *description,
		      int (*fn)(int, int, loff_t, int))
{
	loff_t offset;
	int rc, block = 0, errors = 0, ignored;
	
	/* Round the start position down to the nearest step */
	block = (start / step);
	offset = block * step;

	while (file_limit < 0 || offset < file_limit) {
		progress("  %s block %d [offset %s]...",
			 description, block++, describe_offset(offset, 0));
		rc = fn(fd, chunksize, offset, counter);
		if (rc) {
			/* Don't count errors at EOF which re-set the
			 * file limit */
			ignored = (file_limit > 0 && offset >= file_limit);
			if (!ignored)
				errors++;
			flush_console();
			if (rc > 0) {
				fprintf(stderr, 
					"Error %s file at offset %jd (%s): "
					"%s%s\n",
					description, offset, 
					describe_offset(offset, 0), strerror(rc),
					ignored ? " (EOF, ignored)" : "");
				break;
			} else {
				fprintf(stderr, 
					"    Data error at offset %jd, "
					"continuing.\n",
					offset);
			}
		}
		offset += step;
	}
	if (!quiet) {
		flush_console();
		printf("  Completed %s file with %d errors.\n", 
		       description, errors);
	}
	return errors;
}

int do_writes(int fd, loff_t start, int chunksize, loff_t step, int counter)
{
	return iterate_over_file(fd, start, chunksize, step, counter,
				 "writing", write_one);
}

int do_reads(int fd, loff_t start, int chunksize, loff_t step, int counter)
{
	return iterate_over_file(fd, start, chunksize, step, counter, 
				 "reading", read_one);
}

	
int main(int argc, char *argv[])
{
	int	rc, fd, ch;
	char	*filename;
	int	chunksize = default_chunksize;
	loff_t  step = default_step;
	loff_t	size, start = 0;
	int	i, cycles = 1;
	int	only_read = 0, only_write = 0, one_cycle = 0;
	int	append = 0, query = 0, do_nothing = 0;
	void	*tmp;
	
	progname = argv[0];

	while ((ch = getopt(argc, argv, "1ahQVqrws:")) != EOF) {
		switch (ch) {
		case '1':
			one_cycle++;
			break;
		case 'a':
			append++;
			break;
		case 'V':
			show_version();
			do_nothing++;
			break;
		case 'h':
			show_help();
			do_nothing++;
			break;
		case 'Q':
			query++;
			break;
		case 'q':
			quiet++;
			break;
		case 'r':
			only_read++;
			break;
		case 'w':
			only_write++;
			break;
		case 's':
			chunksize = memparse(optarg);
			break;
		default:
			useage(1);
		}
	}
	argc -= optind;
	argv += optind;
	
	if (do_nothing)
		exit(0);
	
	if (argc < 1 || argc > 3)
		useage(1);

	if (only_read && only_write) {
		fprintf(stderr, "-r and -w are mutually exclusive\n");
		useage(1);
	}
	if (only_read || only_write)
		one_cycle = 1;
	
	filename = argv[0];
	if (argc >= 2) 
		step = memparse(argv[1]);
	if (argc >= 3)
		cycles = memparse(argv[2]);
	
	if (step < chunksize)
		step = chunksize;

	rc = posix_memalign(&tmp, getpagesize(), chunksize);
	if (rc)
		die("Unable to allocate %d bytes for IO buffer.\n");
	buffer = (char *) tmp;

	/* Make sure we don't die with SIGXFSZ if we go beyond the size
	   limit for a regular file: we want to catch the EFBIG error
	   normally. */
	signal(SIGXFSZ, SIG_IGN);

	/* Progress logging stuff wants to know if output is to a tty */
	progress_tty = isatty(1);
	
	fd = open64(filename, O_RDWR | O_DIRECT);
	if (fd < 0)
		die_perror("Failed to open file \"%s\"", filename);

	query_state(fd, &size, query);
	if (query)
		return 0;
	if (append)
		start = size;
	
	i = 0;
	do {
		int errors_this_cycle = 0;
		
		i++;
		if (one_cycle)
			i = cycles;
		
		if (cycles != 1 && !quiet) 
			printf("***\n"
			       "***Beginning write/read cycle %d\n"
			       "***\n", i);

		if (!only_read) {
			if (!quiet)
				printf("Writing for %s bytes every %s bytes\n", 
				       describe_offset(chunksize, obuf1), 
				       describe_offset(step, obuf2));
			errors_this_cycle += do_writes(fd, start,
						       chunksize, step, i);
		}
		
		if (!only_write) {
			if (!quiet)
				printf("%s for %s bytes every %s bytes\n", 
				       only_read ? "Reading" : "Rereading",
				       describe_offset(chunksize, obuf1), 
				       describe_offset(step, obuf2));
			errors_this_cycle += do_reads(fd, start,
						      chunksize, step, i);
		}
		
		flush_console();
		
		if (errors_this_cycle) {
			flush_console();
			fprintf(stderr, "Aborting cycle %d after %d errors\n",
				i, errors_this_cycle);
			break;
		}
	} while (i < cycles);
	
	return 0;
}
